#! /usr/bin/env python
# -*- Mode: Python -*-
# -*- coding: ascii -*-

__author__	   = "AutoGui, Steven Pettit"
__date__	   = "05 Aug 2018"
__copyright__  = "AutoGui, Steven Pettit"
__version__	   = "1.0"
__maintainer__ = "Steven Pettit"
__email__	   = "kryslin@tlhea.org"
__status__	   = ""
__lwver__	   = "11"

import lwsdk

class MultitreeNode(object):
	def __init__(self, name_list):
		self.name_list = name_list
		self.flags = 0
		self.children = []

class Gui(lwsdk.ICommandSequence):
	def __init__(self, context):
		super(Gui, self).__init__()

		self.mylist = []
		self.row_select = []

	def stylemaps_name_2d(self, control, userdata, row, a):
		if row == -1: return "stylemaps"
		else: return self.stylemaps_listNames[row]

	def stylemaps_count_2d(self, control, userdata):
		return len(self.stylemaps_listNames)

	def stylemaps_multi_select_event_func(self, control, user_data, row, selecting):
		if row < 0: return	# list selections are being cleared
		action = "deselected"

		if selecting:
			#print "Selected Row", row
			action = action[2:]
			self.row_select[row] = True
		else:
			#print "De-Selected Row", row
			self.row_select[row] = False

	def stylemaps_null(self, control, a, b):
		return

	def process(self, mod_command):
		def callback(id, data):
			pass #Remove this line if there are other callbacks
		
		self.objf = lwsdk.LWObjectFuncs()
		self.mapObj = self.objf.numVMaps(lwsdk.LWID_('GUDE'))
	
		if not self.mapObj:
			lwsdk.LWMessageFuncs().error("No Style Maps in Object!","")
			return lwsdk.AFUNC_OK
	
		for i in range(0,self.mapObj):
			self.mylist.append(self.objf.vmapName(lwsdk.LWID_('GUDE'),i))
		
		self.ui	   = lwsdk.LWPanels()
		self.panel = self.ui.create("Delete Style Maps 1.0")
		
		# In order to add items to the list, this is how you do it, and then refresh the control:
		# just add elements to the array: self.stylemaps_listNames
		
		self.stylemaps_listNames = self.mylist
		self.row_select = [False for t in self.mylist]
		self.stylemaps = self.panel.multilist_ctl("stylemaps", 370, 10, self.stylemaps_name_2d, self.stylemaps_count_2d, self.stylemaps_null)
		self.stylemaps.set_select(self.stylemaps_multi_select_event_func)
		
		self.panel.align_controls_vertical([self.stylemaps])
		self.panel.size_to_layout(5, 5)
		
		if self.panel.open(lwsdk.PANF_BLOCKING | lwsdk.PANF_CANCEL) == 0:
			self.ui.destroy(self.panel)
			return lwsdk.AFUNC_OK # when cancel was clicked

		self.selected = []
		for i in range(len(self.mylist)):
			if self.row_select[i]:
				self.selected.append(self.mylist[i])

		if len(self.selected):
			
			meo = mod_command.editBegin(0,0,lwsdk.OPSEL_USER)
			
			if not meo:
				print ("Failed to init Mesh Edit Operations...")
				return lwsdk.AFUNC_OK
			else:
				for i in range(len(self.selected)):
					vmid = meo.vMapSelect(meo.state, self.selected[i], lwsdk.LWID_('GUDE'),9)
					if vmid:
						meo.vMapRemove(meo.state)
				
			meo.done(meo.state, lwsdk.EDERR_NONE,0)
			
		return lwsdk.AFUNC_OK

ServerTagInfo = [("DelStyleMaps", lwsdk.SRVTAG_USERNAME | lwsdk.LANGID_USENGLISH ), ("Delete Style Maps", lwsdk.SRVTAG_BUTTONNAME | lwsdk.LANGID_USENGLISH )]
ServerRecord = { lwsdk.CommandSequenceFactory("DelStyleMaps", Gui) : ServerTagInfo }
