@warnings
@name Style Comb

//Basic Style Combing for Fiber FX
//Major Version : 2 (New Algorithm Update)
//Minor Version : 0 (Bug Fixes, Functionality)
//Author : Steven "Kryslin" Pettit 17 July 2018
//Updated 24 Feb 2020

//Global Variables...
//-->
//FiberFx Guide VMap
@define VMAP_GUDE @'G','U','D','E'@
//Number of Dimensions : 3 Vectors, for a total of 9 entries!

//Interface variables
var c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11,c12,c13,c14,c15,c16,c17,c18;
var c99,c98;

//Listbox and Working surfaces list
var lb_items;
var pb_items;

//master surface and parts list
var m_surfs;
var m_parts;
var parts_flag = true;

var m_bb =@<0,0,0>,<0,0,0>@;

var state_flag = false;

var usr_mode = 0; //Default mode, use surf/part to brush away from (0); Arbitrary Point Selection is (1)
var usr_srf = 0;  //Use Surfaces (0) instead of parts (1);
//<--
main {
//Variable Initialization
//-->	
	var VERSION = "2.01";
	var away_s = <0,0,0>;
	var mapname = "style";
	var blend_strength = 1.00;
	
	var bbox1 = @<0,0,0>,<0,0,0>@;
	var bbox2 = @<0,0,0>,<0,0,0>@;
	var tmp = @<0,0,0>,<0,0,0>@;

	var guide = @<0,0,0>,<0,0,0>,<0,0,0>@;
	var fib_length = getFFXlength();
	var axs = 0.00;
	var pc_id;
	var pc_bv;
	
	var wmapnames = nil;
	var widx = @1,1,1@;

	m_bb = boundingbox();
	
	var srf_idx = 1;
	var prt_idx = 1;
	var gs = 0;
	var x_axis = false;
	var z_axis = false;
	var parts_flag = true;
	
	var crl = 0.15;
	var pp[2];
	
//<--
//Initialize Arrays, Trap Conditions
//-->
	
	//Get the part names!
	m_parts = getpartnames();
	//Need this to get surface names
	m_surfs = getsurfnames();
	//Let's get GUDE maps...
	var tmp = getstylemaps();
	var my_style = tmp[2];
	style_idx = tmp[1];
	//Get Weight Map Names;
	wmapnames = getweightmaps();
	//Let's check for what kind of selection has been made
	usr_mode = points_selected();
	//If there are points selected, grab the IDs now.
	if(usr_mode){
		editbegin();
			tmp_v = points;
		editend(ABORT);
	}
	//Let's trap symmetry;  it appears to negatively affect performance.
	if(getmode("SYMMETRY")) error("Please turn off symmetry, and re-run script.");
	//Let's check for possible degenerate geometry...
	if(degengeocheck()) error("This mesh has degenerate geometry.  Please fix before running this script again.");
	/*Let's trap no parts defined as well...
		This currently does nothing, but it should
		disable the ability to select parts and load
		the listbox, to avoid errors.*/
	if ((m_parts.count() < 2) || (m_parts.isNil())) parts_flag = false;
//<--
//Interface
//-->
	selmode(USER);
	reqbegin("Style Comb " + VERSION);
		reqsize(272,584);
		//Let's be tricky...
		//This is a flag to hide a control should a point selection be made
		//There will be one for if there are no parts defined as well.
		//-->
		c99 = 	ctlcheckbox("Hidden",usr_mode);
				ctlposition(c99,8,-99);
		c98 = 	ctlcheckbox("Hidden 2",parts_flag);
				ctlposition(c98,8,-75);
		//<--
		//Let's get an existing style map or make a new one
		//-->
		c6 =  	ctlpopup("Style Map",style_idx,my_style);
				ctlposition(c6,8,4);
		c7 =  	ctlstring("New Style Map Name",mapname);
				ctlposition(c7,8,28);
				ctlactive(c6,"bm_visible",c7);		
		//<--
		//Sleekness Control
		//-->
		c8 =  	ctlpopup("Sleekness Control",widx[1],wmapnames);
				ctlposition(c8,8,52);
		c1 =  	ctlpercent("Sleekness",blend_strength);
				ctlposition(c1,8,76);
				ctlactive(c8,"bm_visible",c1);
		//<--
		//Curling
		//-->
		c17 = 	ctlpopup("Curling Control",widx[3],wmapnames);
				ctlposition(c17,8,100);
		c18 =	ctlpercent("Curling",crl);
				ctlposition(c18,8,124);
				ctlactive(c17,"bm_visible",c18);
		//<--
		//Gravity and Axis Bias!!
		//-->
		c13 = 	ctlpercent("Gravity Strength:",gs);
				ctlposition(c13,8,148);
		c14 = 	ctlpercent("Axis Bias Strength:", axs);
				ctlposition(c14,8,172);
		c15 = 	ctlcheckbox("X Axis:",x_axis);
				ctlposition(c15,8, 196);
		c16 =	ctlcheckbox("Z axis:",z_axis);
				ctlposition(c16,128,196);
		//<--
		//Guide Length Control
		//-->
		c9 =  	ctlpopup("Fiber Length Control",widx[2],wmapnames);
				ctlposition(c9,8,220);
		c4 =  	ctldistance("Base Fiber Length",fib_length);
				ctlposition(c4,8,244);
		//<--
		//Get Surfaces or Parts
		//-->
		c10 = 	ctlstate("Use Parts",usr_srf,100,"state_change");
				ctlposition(c10,8,268);
				ctlactive(c10,"cm_visible", c98);
		//<--
		//Surfaces to brush away from 
		//-->
		c2 =  	ctlpopup("Comb away from:" ,srf_idx, m_surfs);
				ctlposition(c2,8,292);
				ctlvisible(c10,"vis1",c2);
				ctlactive(c99,"cm_visible",c2);
		//<--
		//Surfaces or Parts to brush
		//-->
		if(usr_mode){
			//If in Use Point Selection mode, make all surfs/parts available
			srf_idx = 0;
			prt_idx = 0;
		} 
		c5 =  	ctllistbox("Comb these:",240,240, "lb_count","lb_name");
				ctlposition(c5,8,316);
				ctlvisible(c10,"vis1",c5);
				lb_load_items(srf_idx);
				ctlrefresh(c2,"lb_refresh");
		if (parts_flag){
			//Parts to brush away from:
			c11 = 	ctlpopup("Comb away from:", prt_idx, m_parts);
				ctlposition(c11,8,292);
				ctlvisible(c10,"vis2",c11);
				ctlactive(c99,"cm_visible",c11);
			//Parts to Brush
			c12 = 	ctllistbox("Comb these:",240,240, "pb_count","pb_name");
				ctlposition(c12,8,316);
				ctlvisible(c10,"vis2",c12);
				pb_load_items(prt_idx);
				ctlrefresh(c11,"pb_refresh");
		}
		//<--
		requpdate();
		
		return if !reqpost();
		
		blend_strength = getvalue(c1);
		usr_srf = getvalue(c10);
		if(!usr_srf) {
			srf_idx = getvalue(c2);
			test = getvalue(c5);
		} else {
			prt_idx = getvalue(c11);
			test = getvalue(c12);
		}
		fib_length = getvalue(c4);
		style_idx = getvalue(c6);
		if (style_idx == 1){
			mapname = getvalue(c7);
		} else {
			mapname = my_style[style_idx];
		} 
		widx[1] = getvalue(c9);
		widx[2] = getvalue(c8);
		gs = getvalue(c13);
		axs = getvalue(c14);
		x_axis = getvalue(c15);
		z_axis = getvalue(c16);
		widx[3] = getvalue(c17);
		crl = getvalue(c18);
	reqend();	
//<--
//Select a brushing point
//All combing is away from this point

//-->
	if (usr_mode){
		//A point or points have been selected
		selpoint(CLEAR);
		selpoint(SET,POINTID,tmp_v);
	} else {
		//normal surface/part selection
		selpolygon(CLEAR);
		
		if(!usr_srf){
			selpolygon(SET,SURFACE,m_surfs[srf_idx]);
		} else {
			selpolygon(SET,PART,m_parts[prt_idx]);
		}
	}
	bbox1 = boundingbox();
	away_s = center(bbox1);
//<--
//-->		
	//Get surfaces/parts to comb
	selpolygon(CLEAR);
	foreach(v,test){
		if(!usr_srf){
			selpolygon(SET,SURFACE,lb_items[v]);
		} else {
			selpolygon(SET,PART,pb_items[v]);
		}
	}
	bbox2 = boundingbox();
	mcp =center(bbox2);
	bbs = bbox2[2] - bbox2[1];	
	
	td1 = vmag(mcp - <away_s.x,away_s.y, bbox1[1].z>);
	td2 = vmag(mcp - <away_s.x, away_s.y, bbox1[2].z>);
	away_s = <away_s.x,away_s.y, bbox1[1].z>;
	if(td2 > td1) away_s = <away_s.x, away_s.y, bbox1[2].z>;

//<--
//-->	
	undogroupbegin();
	//Safely store these...
	var mbs = blend_strength;
	var mfl = fib_length;
	var mcrl = crl;
	//Drop into Meshedit mode and grab the vertex count.
	moninit(editbegin(),"Combing...");
		//Set up the Style Map
		//-->
		if(style_idx == 1){
			mymap = VMap(VMAP_GUDE, mapname, 9);
		} else {
			mymap = VMap(VMAP_GUDE, mapname);
		}
		if (mymap == nil) error("Style Map " + mapname + " not instanced! "); 
		//<--
		//Have the weight map agents ready for later...
		//-->
		if (widx[1] > 1){
			flmap = VMap(VMWEIGHT,wmapnames[widx[1]]);
			if (flmap == nil) error("Length Map " + wmapnames[widx[1]] + " not instanced! "); 
		} 
		if (widx[2] > 1){ 
			slmap = VMap(VMWEIGHT,wmapnames[widx[2]]);
			if (slmap == nil) error("Sleek Map " + wmapnames[widx[2]] + " not instanced! "); 	
		}
		if(widx[3] > 1){
			clmap = VMap(VMWEIGHT,wmapnames[widx[3]]);
			if(clmap == nil) error("Curling Map " + wmapnames[widx[3]] + "not instanced!");
		}
		//<--		
		foreach(vtx,points){
			//Get the polygon infomation for the point...
			p = nil;
			p = vtx.polygon();
			//Trap points with no polygons, don't process them			
			if(p != nil) {
				pt = pointinfo(vtx);
				tmp = make_vtx_perp(vtx, away_s);
				vn = tmp[1];
				vb = tmp[2];
				//Gravity and Axis Bias
				m_bias = make_master_bias(pt, vn, away_s,bbs, x_axis, z_axis);
				//The effect of axis bias on the perpendicular
				tv = <sgn(pt.x - away_s.x) * m_bias.x,0,sgn(pt.z - away_s.z) * m_bias.z >;
				vb += (tv * axs);
				//Keep everything nice...
				vn = normalize(vn);
				vb = normalize(vb);
				//And get the perpendicular vector
				vb = makeperp(vb,vn);
				//Now, let's do our weightmap stuff...
				//Recall the master values...
				fib_length = mfl;
				blend_strength = mbs;
				crl = mcrl;
				//Fur Length
				if(widx[1] > 1){
					tv = nil;
					if (flmap.isMapped(vtx)){
						tv = flmap.getValue(vtx);
						fib_length = mfl  * max(0,tv[1]);	
					} else {
						fib_length = 0.0;
					}
				}
				//Sleekness
				if(widx[2] > 1){
					tv = nil;
					if (slmap.isMapped(vtx)) {
						tv = slmap.getValue(vtx);
						blend_strength = tv[1];
					} else {
						blend_strength = 0.0;
					}
				}
				//Fiber curl
				if(widx[3] > 1){
					tv = nil;
					if(clmap.isMapped(vtx)){
						tv = clmap.getValue(vtx);
						crl = tv[1];
					} else {
						crl = mcrl;
					}
				}
				//Gravity cancels out sleekness, curl
				blend_strength = (1 - m_bias.y) * blend_strength;
				crl = (1 - m_bias.y) * crl;
				//Make our guide.
				guide = makeguide2(pt, vn, vb, fib_length, blend_strength, crl);
				//convert it to an array
				ta = makearray(guide);
				//And write it to the style map
				mymap.setValue(vtx, ta);
			}
		monstep();
		}
	monend();
	editend();
	
	undogroupend();
	selpolygon(CLEAR);
//<--
}

make_master_bias :e_pt, e_vn, b_pt,t_bb, u_x, u_z
{
	//Master bias routine
	//returns a vector, each field containing a value of 0..1
	//which controls how much the each axis (X, -Y, Z) affect
	//the combing of the mesh.
	var rv = <0,0,0>;
	var c = <0,0,0>;
	var g, xd, zd;
	//First up is gravity
	//affect of gravity is based on slope of the surface normal
	//the closer to <0,-1,0> it is, the stronger the effect is.
	//This will be applied to the sleekness control, so as to not brush
	//vertical hairs
	//g = (((e_vn.y * -1.0) * 0.5) +0.5) ^^ 2;
	rv.y = clamp(dot3d(e_vn,<0,-1,0>),0,1) ^^ 2;
	//next up is x and z axis bias
	//if neither x or z axis is needed, we'll skip.
	if(u_x || u_z){
		//get evaluation vector
		c = e_pt - b_pt;
		//get axis distances
		//If either axis distance = 0, don't do the math, just return 0.
		if (u_x && (t_bb.x != 0)) rv.x = abs(c.x)/(t_bb.x / 2);
		if (u_z && (t_bb.z != 0)) rv.z = c.z/(t_bb.z);
	}
	//return the values
	return(rv);
}

getpartnames
{
	/* This is brute force, but non_desctructive.
		Get Part Names! */
	var retv, p;
	selmode(GLOBAL);
	editbegin();
		foreach(p,polygons){
			pn = p.tag("PART");
			if (retv){
				q = retv.indexOf(pn);
				if(!q) retv += pn;
			} else {
				retv += pn;
			}
		}
	editend(ABORT);
	return(retv);
}

getsurfnames
{
	var retv = Surface(Mesh());
	return(retv);
}

points_selected
{
	//get any points directly selected
	selmode(DIRECT);
	var retv = (pointcount() > 0);
	return(retv);
}

degengeocheck
{
	var retv = false;
	//Let's check for possible degenerate geometry...
	tmp = lyrfg();
	fgl = tmp[1];
	tmp = lyrempty();
	mtl = tmp[1];
	selmode(USER);
	selpolygon(CLEAR);
	selpolygon(SET);
	copy();
	lyrsetfg(mtl);
	paste();
	mergepoints();
	tmp = polycount();
	//Let's clean up our mess...
	delete();
	lyrsetfg(fgl);
	//And return true if there are any one or two point polys found
	if ((tmp[2] >0) || (tmp[3] >0)) retv = true;
	return(retv);
}

getstylemaps
{
	var retv[2];
	var tmp = VMap(VMAP_GUDE);
	var tsa = nil;
	var tsi = 2;
	tsa += "(New)";
	qt = 0;
	if (tmp) qt = tmp.type;
	while(tmp && tmp.type == -1)
	{
		if (tmp.dimensions == 9) tsa += tmp.name;
		tmp = tmp.next();
	}
	if (tsa.count() == 1) tsi == 1;
	retv[1] = tsi;
	retv[2] = tsa;
	return(retv);
}

getweightmaps
{
	retv = nil;
	tmp = VMap(VMWEIGHT);
	retv += "Fixed";
	while(tmp && tmp.type == VMWEIGHT)
	{
		retv += tmp.name;
		tmp = tmp.next();
	}
	return(retv);
}

sgn : a
{
	var retv = 0;
	if (a > 0) retv = 1;
	if (a < 0) retv = -1;
	return(retv);
}

getFFXlength
{
	var d;
	var tb = boundingbox();
	var ub = tb[2]-tb[1];
	d = sqrt(max(ub.x^^2 + ub.y^^2, max(ub.y^^2 + ub.z^^2,ub.z^^2 + ub.x^^2))) / 20.0;
	return(d);
}

makearray : Vecs
{
	var retv[9];
	retv[1] = Vecs[1].x;
	retv[2] = Vecs[1].y;
	retv[3] = Vecs[1].z;
	retv[4] = Vecs[2].x;
	retv[5] = Vecs[2].y;
	retv[6] = Vecs[2].z;
	retv[7] = Vecs[3].x;
	retv[8] = Vecs[3].y;
	retv[9] = Vecs[3].z;
	return(retv);
}

makeperp : vector_p, vector_q
{
	var retv = <0,0,0>;
	var v_p = normalize(vector_p);
	var v_q = normalize(vector_q);
	//variation on perpendicular projection;
	//since v_q is normalized, ||v_q|| = 1, ||v_q||^2 = 1;
	//Let's see if my assumption is correct...
	retv = normalize(v_p - (dot3d(v_p,v_q) * v_q));
	if (vmag(retv) == 0) retv = v_p;
	return(retv);
}

make_vtx_perp : p_eval, q_brush
{
	//Take a point, generates a vertex normal, and computes an arbitrary perpendicular
	//based on simple computations.
	//p_eval = point to be evaluated;
	//q_brush = point of orgin for brushing vector
	var retv = @<0,0,0>,<0,0,0>@;
	var bp = pt = tv = vn = rv = xz= <0,0,0>;
	//Get the point info and brushing vector;
	pt = pointinfo(p_eval);
	bp = q_brush;
	tv = normalize(pt - bp);
	//Get the vertex normal
	var pc = p_eval.polygon;
	var pids = p_eval.polygon();
	foreach(p,pids) vn += polynormal(p) / pc;
	//The programmer is an idiot, who could have used this much earlier.
	//Vector projected onto a plane defined by a surface normal
	rv = normalize(tv - (dot3d(vn,tv) * vn));
	//A couple of quick checks...
	vn = normalize(vn);
	rv = normalize(rv);
	if(((pt.y + rv.y) > (pt.y + vn.y)) && (vn.z > 0)) rv *= <1,-1,1>;
	//check to see if the perp is pointing towards the brush point on the x axis, instead of away...
	//if( dot3d(rv,normalize(<-pt.x,0,0>)) > 0) rv.x = -rv.x;
	if(sgn(rv.x) != sgn(pt.x)) rv.x = -rv.x;
	//Normalize the lot of them
	//Vertex Normal
	retv[1] = normalize(vn);
	//Vertex Perpendicular
	retv[2] = normalize(rv);
	//return the vertex normal and generated perpendicular.
	return(retv);
}

makeguide2 : Porig, Vnorm, Vbrush, Flen, Sleek, curl
{
	var retv = @<0,0,0>,<0,0,0>,<0,0,0>@;
	var ps[2][2];
	var ts = clamp(Sleek,0,1);
	var nx = ny = theta = 0.0;
	var lcrl = curl;
	
	ps[1][1] = Porig;
	ps[2][1] = Porig + (Flen * Vbrush);
	ps[1][2] = ps[1][1] + (Flen * Vnorm);
	ps[2][2] = ps[2][1] + (Flen * Vnorm);
		
	theta = clamp(acos(ts),rad(10),rad(90));
	retv[1] = lerp2d(ps,0,0);
	
	lcrl = clamp(curl,0,1);
	lcrl = rad(10) + lcrl * rad(60);
		
	for(i=0;i<2;i++){
		nx += 0.5 * cos(theta);
		ny += 0.5 * sin(theta);
		retv[i+2] = lerp2d(ps,nx,ny);
		theta -= lcrl;
		if (theta < 0) theta += 2 * lcrl;
	}
	return(retv);
}

lerp2d : A, T, U 
{
	retv = <0,0,0>;
	//A contains a 2 x 2 array of XYZCoords
	a = A;
	t = clamp(T,0,1);
	u = clamp(U,0,1);
	wrkp1 = lerp(a[1][1],a[1][2],u);
	wrkp2 = lerp(a[2][1],a[2][2],u);
	retv = lerp(wrkp1,wrkp2,t);
	return(retv);
	
}

lerp : A, B, T 
{
	retv = <0,0,0>;
	retv = A + (B - A) * T;
	return(retv);
}

slerp : p0, p1, t 
{
	retv = <0,0,0>;
	var v0 = normalize(p0);
	var v1 = normalize(p1);
	var it = <1 - t, 1 - t, 1 - t>;
	var ot = <t,t,t>;
	var dp = dot3d(v0,v1);
	if (dp < 0){
		v1 = -v1;
		dp = -dp;
	}
	if(dp > 0.995){
		//To close for comfort, linear interpolate and normalize
		retv = normalize((it * p0) + (ot * p1));
	} else {
		th0 = acos(dp);
		th1 = th0 * t;
		sin_th1 = sin(th1);
		sin_th0 = sin(th0);
		s0 = cos(th1) - dp * sin_th1 / sin_th0;
		s1 = sin_th1 / sin_th0;
		retv = (s0 * v0) + (s1 * v1);
	}
	return(retv);
}

lb_count
{
	return(lb_items.count());
}

lb_name: index
{
	return(lb_items[index]);
}

lb_load_items : idx
{
	lb_items = nil;
	for(i=1;i<=m_surfs.count();i++) if (i != idx) lb_items += m_surfs[i];
	setvalue(c5,lb_items.count());
}

lb_refresh : value
{
	lb_load_items(value);
	requpdate();
}

pb_count
{
	return(pb_items.count());
}

pb_name: index
{
	return(pb_items[index]);
}

pb_load_items : idx
{
	pb_items = nil;
	for(i=1; i<=m_parts.count();i++) if(i != idx) pb_items += m_parts[i];
	setvalue(c12,pb_items.count());
}

pb_refresh : value
{
	pb_load_items(value);
	requpdate();
}

bm_visible : value
{
	retv = 0;
	if (value == 1) retv = 1;
	return(retv);
}

cm_visible : value
{
	return(!value);
}

state_change : value
{
	state_flag = true;
}

vis1 : value
{
	return(!value);
}

vis2 : value
{	 
	return(value && parts_flag);
}
